## @package ardi.api.assets
#  ARDI Asset API
#
#  These classes allow you to interrogate and modify assets.
import pycurl
import urllib
import xmltodict
try:
    from io import BytesIO
except ImportError:
    from StringIO import StringIO as BytesIO

## Represents a single ARDI asset
#
#  This class represents a single ARDI asset
#
#  @todo This class is a placeholder and has little-to-no functionality as yet.
class ARDIAsset:
    def __init__(self, core, id):
        self.core = core
        self.id = id
        self.name = ""

    #Downloads the name and other identifying information of this asset.
    def Identify(self):
        xml = self.core.XMLAPICall("api/asset/id",{ 'id' : self.id })
        if xml != False:
            #print(str(xml))
            try:
                self.name = xml['response']['asset']['name']
                return True
            except:
                pass
        return False

    def GetData(self,includelive=True):
        resp = []

        #final = urllib.urlencode({ 'id' : self.id})

        #buffer = BytesIO()
        #ar = pycurl.Curl()
        #ar.setopt(ar.URL, 'http://' + self.core.server + ':' + str(self.core.port) + '/api/getvalues.php?' + final)
        #ar.setopt(ar.WRITEFUNCTION, buffer.write)
        #ar.perform()
        
        # HTTP response code, e.g. 200.
        #if ar.getinfo(ar.RESPONSE_CODE) == 200:
        xml = self.core.XMLAPICall("api/asset/info",{ 'id' : self.id, 'live' : includelive })
        if xml != False:
            #print(str(xml))
            res = xml['asset']['property']
            for ele in res:
                vl = { 'id': ele['@id'], 'name' : ele['@name'], 'type' : ele['@type'],'origin':ele['@origin'],'origintype':ele['@origintype'], 'value' : {} }
                for key in ele.keys():
                    if key[0] != '@':
                        vl['value'][key] = ele[key]
                
                resp.append(vl)
            
        #ar.close()
        return resp
        #pass

    def GetEffects(self):
        xml = self.core.XMLAPICall("api/asset/effects",{ 'id' : self.id })
        if xml != False:
            return xml['behaviours']
            
        return False

    def GetRelated(self):
        resp = []
        
        xml = self.core.XMLAPICall("api/asset/getrelationships",{ 'id' : self.id })
        if xml != False:
            #print(str(xml))
            res = xml['response']['relatedto']
            #if type(res) != type([]):
            #    res = (res,)
                
            for ele in res:
                #print(str(ele))
                vl = { 'id': ele['@asset'], 'name' : ele['@name'], 'rel' : ele['@rel'],'dir':ele['@dir']}
                resp.append(vl)
            
        return resp

    def GetAllRelatedBy(self,relid,flat=True):
        resp = []
        
        xml = self.core.XMLAPICall("api/asset/getrelationships",{ 'id' : self.id, 'indirect' : 'yes', 'rel': relid })
        if xml != False:
            res = xml['response']['relatedto']
            if not isinstance(res,list):
                res = (res,)
                
            if flat == True:
                for ele in res:
                    vv = ele
                    while True:
                        vl = { 'id': vv['@asset'], 'name' : vv['@name'], 'rel' : vv['@rel'],'dir': vv['@dir'],'dist': vv['@distance']}
                        resp.append(vl)
                        try:
                            vv = vv['relatedto']
                        except:
                            break
            else:
                for ele in res:
                    vv = ele
                    while True:
                        vl = { 'id': vv['@asset'], 'name' : vv['@name'], 'rel' : vv['@rel'],'dir': vv['@dir'],'dist': vv['@distance']}
                        resp.append(vl)
                        try:
                            vv = vv['relatedto']
                        except:
                            break
        return resp

    def ValueToString(self,ele,typename):
        valuetext = ""
        if typename == 'TAG':
            return ele['text']
        if typename == 'TEXT':
            return ele['text']
        if typename == 'MEASUREMENT':
            try:
                return ele['measurement'] + ele['units']
            except:
                if ele['measurement']['#text'] == '^':
                    return "[ERROR]"
                return ele['measurement']['#text'] + ele['units']

        if typename == 'STATUS':
            vl = ele['state']
            if vl == "1":
                vl = ele['on']
            else:
                vl = ele['off']
            return vl
        if typename == 'ENUM':
            try:
                v = int(ele['value'])
            except:
                if ele['value']['#text'] == '^':
                    return "[ERROR]"
                v = int(ele['value']['#text'])

            trans = ele['enum'].split('^')
            for e in trans:
                parts = e.split('=')
                
                bits = parts[0].split('-')
                #print "Checking " + bits[0]
                if len(bits) == 1:
                    if v == int(bits[0]):
                        return parts[1]
                else:
                    if (v >= int(bits[0])) and (v <= int(bits[1])):
                        return parts[1]
            return "UNKNOWN"

        if typename == 'LOOKUP':
            if ele['text'] is None:
                return "None"
            try:
                return ele['text']['#text']
            except:
                return ele['text']
        return "[UNKNOWN]"

## Set asset properties and values
#
#  This class is used to get and set properties on ARDI assets.
#
class ARDIAssets:
    ## Default Constructor
    #
    #  @param core The ARDICore object used to connect to the ARDI server.
    def __init__(self, core):
        self.core = core

    ## Requests a property value across all or a subset of assets
    #
    #  This function provides the value of one or more properties across multiple assets.

    def GetPropertyValues(self,properties,assets=[]):
        resp = []

        params = { 'props' : ','.join(properties)}
        if len(assets) != 0:
            params['assets'] = ','.join(assets)
        
        xml = self.core.XMLAPICall("api/asset/propvalue",params)
        if xml == False:
            return False

        try:
            rx = xml['results']['property']
            #if not isinstance(xml['results']['property'],dict):
            #    print 'Fudging!'
            #    rx = [rx]
        except:
            return False

        for n in rx:
            q = {}
            for r in n:
                #print(str(r))
                if r[0] != '@':
                    q[r] = n[r]
            elem = {'asset': n['@asset'], 'property': n['@id'], 'value': q}
            resp.append(elem)
        #print(str(xml))
        return resp

    ## Sets an individual property node
    #
    #  This function sets an individual property node in an ARDI asset.
    #
    #  For example, SetInput(15,24,'measurement',20) would change the 'measurement' of property 24 on asset 15 to '20'.
    def SetInput(self,asset,property,input,value):

        final = urllib.urlencode({ 'asset' : asset, 'property': property, 'input' : input, 'value' : value })

        resp = []

        buffer = BytesIO()
        ar = pycurl.Curl()
        ar.setopt(ar.URL, 'http://' + self.core.server + ':' + str(self.core.port) + '/api/setdata.php?' + final)
        ar.setopt(ar.WRITEFUNCTION, buffer.write)
        ar.perform()
        
        # HTTP response code, e.g. 200.
        if ar.getinfo(ar.RESPONSE_CODE) == 200:    
            xml = xmltodict.parse(buffer.getvalue())

            ar.close()

            #print xml;
            try:
                res = xml['response']['success']
                return True
            except:
                return False            
            
        ar.close()
        return False
    ## Sets all nodes of a property
    #
    #  This function sets a full property on an ARDI asset.
    #
    #  For example, SetInput(15,24,'20| Deg C|0|100|2') would (assuming it's a measurement) set property 24 on asset 15 to '20 Deg C' in the range between 0-100 and with 2 decimal places of accuracy.
    def SetData(self,asset,property,value):

        final = urllib.urlencode({ 'asset' : asset, 'property': property, 'value' : value })

        resp = []

        buffer = BytesIO()
        ar = pycurl.Curl()
        ar.setopt(ar.URL, 'http://' + self.core.server + ':' + str(self.core.port) + '/api/setdata.php?' + final)
        ar.setopt(ar.WRITEFUNCTION, buffer.write)
        ar.perform()
        
        # HTTP response code, e.g. 200.
        if ar.getinfo(ar.RESPONSE_CODE) == 200:    
            xml = xmltodict.parse(buffer.getvalue())

            ar.close()

            #print xml;
            try:
                res = xml['response']['success']
                return True
            except:
                return False            
            
        ar.close()
        return False

    ## Links a property to live data
    #
    #  This function is used to link an existing asset property to live or historical data.
    #
    #  @todo This function is incomplete and not working - it currently calls 'setdata' instead.
    def LinkData(self,asset,property,value,source,address):

        final = urllib.urlencode({ 'asset' : asset, 'property': property, 'value' : value })

        resp = []

        buffer = BytesIO()
        ar = pycurl.Curl()
        ar.setopt(ar.URL, 'http://' + self.core.server + ':' + str(self.core.port) + '/api/setdata.php?' + final)
        ar.setopt(ar.WRITEFUNCTION, buffer.write)
        ar.perform()
        
        # HTTP response code, e.g. 200.
        if ar.getinfo(ar.RESPONSE_CODE) == 200:    
            xml = xmltodict.parse(buffer.getvalue())

            ar.close()

            #print xml;
            try:
                res = xml['response']['success']
                return True
            except:
                return False            
            
        ar.close()
        return False
