## @package ardi.api.core
#  ARDI Server Connection API
#
#  Classes in this package allow you to establish a connection to a site on an ARDI server.
#
#  Once connected, you use classes from the other ARDI namespaces to actually communicate with the server - the classes here are simply intended to allow you to connect.

import pycurl
import urllib
import xmltodict
import requests

try:
    from io import BytesIO
except ImportError:
    from StringIO import StringIO as BytesIO

##Contains the details about a single ARDI context.
#
#The ARDICore object offers an array of these assets - one for each context present in the connected server.
class ARDIContext:
    def __init__(self):
        ##The port number of the consolidator that you request live data from
        self.consolidator = 5336
        ##The port number used by drivers to submit to the consolidator
        self.submission = 2991
        ##The name of the context
        self.name = "Actual"
        ##The name of the server the context is attached to
        self.server = ""

## Manages Connection To And Queries About ARDI Servers.
#
# This object is used to both connect to ARDI servers to verify their properties (via connect()) and to query driver and metadata information.
# Note that all other communication is done through different API functions.
class ARDICore:
    ## Default Constructor
    #
    # Pass the name of the server as part of the constructor for the ARDICore. This
    # is the FULL name of the site ( ie ardi.mycompany.org/s/default )
    #
    # @param srv The name (host name or IP address) of the ARDI server.
    def __init__(self, srv, site="default"):
        ##The name of the ARDI server
        self.server = srv
        ##The site name
        self.site = site
        ##The HTTP port that the server is communicating on
        self.webport = 80
        ##A list of ARDIContext objects representing each context found on the server.
        self.contexts = []

    ## Call an API function...

    def APICall(self,url,params):
        r = requests.get('http://' + self.server + ':' + str(self.webport) + '/s/' + self.site + '/' + url,params)

        #print 'http://' + self.server + ':' + str(self.webport) + '/s/' + self.site + '/' + url + " ( " + str(params) + ")"

        # HTTP response code, e.g. 200.
        if r.status_code == 200:
            content = r.text
            #content.replace("\r","\r\n")
            #ps = content.index('?>')
            #content = content[ps+2:].strip()
            #print "Complete: " + str(content)
            return content
        else:
            print "API Call Failed: " + str(r.status_code)
            return False

    def XMLAPICall(self,url,params):
        r = requests.get('http://' + self.server + ':' + str(self.webport) + '/s/' +self.site + '/' + url,params)

        #print 'http://' + self.server + ':' + str(self.webport) + '/s/' + self.site + '/' + url + " ( " + str(params) + " )"

        # HTTP response code, e.g. 200.
        if r.status_code == 200:
            content = r.text
            #ps = content.index('?>')
            #content = content[ps+2:].strip()
            try:
                xml = xmltodict.parse(content,r.encoding)
            except:
                print "API Call Failed Parsing XML"
                return False
            return xml
        else:
            print "API Call Failed: " + str(r.status_code)
            return False

    ## Connect to the ARDI server and get metadata.
    #
    #  This establishes a connection to ARDI, requesting the list of contexts
    #  available.
    def Connect(self):
        
        content = self.XMLAPICall("api/connect",None)
        if content != False:
            #print 'XML Content: "' + str(content) + '"'
            #xml = xmltodict.parse(content)
            xml = content
            for d in xml['ardi']['service']:
                #print 'Checking ' + d['@name']
                if d['@name'] == 'data':
                    prt = 5336
                    src = self.server
                    try:
                        prt = d['@port']
                    except:
                        pass
                    
                    try:
                        src = d['@host']
                    except:
                        pass                

                    ctx = ARDIContext()
                    ctx.consolidator= prt
                    ctx.server = src
                    ctx.name = "Actual"
                    self.contexts.append(ctx)
            return True
        else:
            print "Server connection failed."
            #ar.close()
            return False        
        

    ## Gets metadata from the server.
    #
    # This function returns a tuple containing a list of both the relationships
    # and properties found on the ARDI server.
    def GetConfiguration(self):
        #buffer = BytesIO()
        #ar = pycurl.Curl()
        #ar.setopt(ar.URL, 'http://' + self.server + ':' + str(self.webport) + '/api/getconfiguration.php')
        #ar.setopt(ar.WRITEFUNCTION, buffer.write)
        #ar.perform()        

        xml = self.XMLAPICall("api/getconfiguration",None)
        if xml != False:

            #print(str(xml))
            rels = []
            props = []

            for ele in xml['config']['relations']['relationship']:
                rels.append({ 'name': ele['@name'], 'id': ele['@id'] })                
            
            for ele in xml['config']['properties']['property']:
                props.append({ 'name' : ele['@name'], 'type' : ele['@type'], 'id' : ele['@id'] })
        else:
            return None
                
        #ar.close()        
        return [rels,props]

    ## Requests information about data aquisition.
    #
    # This function returns a tuple containing details of the drivers and data sources
    # in use by this ARDI server.
    def GetDataSourceInfo(self):
        #buffer = BytesIO()
        #ar = pycurl.Curl()
        #ar.setopt(ar.URL, 'http://' + self.server + ':' + str(self.port) + '/api/getdatasources.php')
        #ar.setopt(ar.WRITEFUNCTION, buffer.write)
        #ar.perform()        

        # HTTP response code, e.g. 200.
        #if ar.getinfo(ar.RESPONSE_CODE) == 200:            
        #   xml = xmltodict.parse(buffer.getvalue())
        xml = self.XMLAPICall("api/getdatasources",None)
        if xml != False:
            drivers = []
            dsources = []

            for ele in xml['config']['drivers']['driver']:
                drivers.append({ 'name': ele['@name'], 'port': ele['@port'], 'code' : ele['@code'] })
            
            for ele in xml['config']['datasources']['source']:
                dsources.append({ 'name' : ele['@name'], 'type' : ele['@type'], 'id' : ele['@id'], 'port' : ele['@port'] })
        else:
            return None
                
        #ar.close()        
        return [drivers,dsources]
