import pycurl
import urllib
import xmltodict
try:
    from io import BytesIO
except ImportError:
    from StringIO import StringIO as BytesIO

## @package ardi.api.search
#  ARDI Search API
#
#  Classes in this package allow you to search for and get properties of ARDI assets.
#
#  @todo This class is currently incomplete and only offers limited functionality.

## Searches ARDI
#
#  You can use this class to perform searches on your ARDI database and get
#  asset information.
class ARDISearch:
    def __init__(self, core):
        self.core = core

    ## Gets a list of assets by Type
    #
    #  This function returns a list of assets based on their type.
    #
    #  It requires the ID number of the type rather than the plain-text type name.
    #
    #  The return value takes the form of an array of dictionaries - for example...
    #
    #  [{ 'id': 22, 'name' : 'Inlet Pressure Transmitter'},{ 'id': 25, 'Outlet Pressure Transmitter'})
    def GetAssetsTaggedAs(self,tag):

        final = urllib.urlencode({ 'taggedas' : tag })

        resp = []

        buffer = BytesIO()
        ar = pycurl.Curl()
        ar.setopt(ar.URL, 'http://' + self.core.server + ':' + str(self.core.webport) + '/api/search.php?' + final)
        ar.setopt(ar.WRITEFUNCTION, buffer.write)
        ar.perform()
        
        # HTTP response code, e.g. 200.
        if ar.getinfo(ar.RESPONSE_CODE) == 200:    
            xml = xmltodict.parse(buffer.getvalue())

            res = xml['results']['result']
            for ele in res:
                resp.append({ 'id': ele['@id'], 'name': ele['@name'] })                
            
        ar.close()
        return resp

    
    ## Searches for an asset
    #
    #  This function returns a list of assets based on the offered search string.
    #
    #  The return value takes the form of an array of dictionaries - for example...
    #
    #  [{ 'id': 22, 'name' : 'Inlet Pressure Transmitter'},{ 'id': 25, 'Outlet Pressure Transmitter'})
    def Find(self,query):

        #final = urllib.urlencode({ 'q' : query })

        resp = []

        xml = self.core.XMLAPICall("api/asset/search",{ 'q' : query })
        if xml != False:
            res = xml['results']['asset']
            if type(res) != type([]):
                res = (res,)
            for ele in res:
                resp.append({ 'id': ele['@id'], 'name': ele['@name'] })                
            
        #ar.close()
        return resp
    
    ## Gets values for an asset
    #
    #  This function requests the values of an asset.
    #
    #  @param assetlist A list of the assets you want to query
    #  @param valuelist A list of the property IDs you want to query.
    def GetAssetValues(self,assetlist, valuelist):

        alist = '|'.join(assetlist)
        vlist = '|'.join(valuelist)       

        final = urllib.urlencode({ 'for' : alist, 'values' : vlist })

        resp = []        

        buffer = BytesIO()
        ar = pycurl.Curl()
        ar.setopt(ar.URL, 'http://' + self.core.server + ':' + str(self.core.port) + '/api/getvalues.php?' + final)
        ar.setopt(ar.WRITEFUNCTION, buffer.write)
        ar.perform()
        
        # HTTP response code, e.g. 200.
        if ar.getinfo(ar.RESPONSE_CODE) == 200:            
            xml = xmltodict.parse(buffer.getvalue())

            res = xml['results']['result']
            for ele in res:
                resp.append({ 'id': ele['@id'], 'property' : ele['@property'], 'value': ele['@value'] })                
            
        ar.close()
        return resp

    ## Gets values for an asset
    #
    #  This function requests the values of an asset.
    def SearchForRelated(self, content, propertyid, relationship):

        mlist = '|'.join(content)       
       
        final = urllib.urlencode({ 'in' : propertyid, 'for' : mlist, 'related' : relationship })

        resp = []

        buffer = BytesIO()
        ar = pycurl.Curl()
        ar.setopt(ar.URL, 'http://' + self.core.server + ':' + str(self.core.port) + '/api/bulksearch.php?' + final)
        ar.setopt(ar.WRITEFUNCTION, buffer.write)
        ar.perform()
        
        # HTTP response code, e.g. 200.
        if ar.getinfo(ar.RESPONSE_CODE) == 200:            
            xml = xmltodict.parse(buffer.getvalue())
                    
            res = xml['results']['result']

            try:                
                resp.append({'id' : res['@id'], 'value' : res['@value'], 'parentname': res['parent']['@name'], 'parentid': res['parent']['@id'] })
            except:
                try:
                    for d in res:                        
                        try:
                            resp.append({'id' : d['@id'], 'value' : d['@value'], 'parentname': d['parent']['@name'], 'parentid': d['parent']['@id'] })
                        except:
                            pass
                except:
                    pass

        ar.close()
        return resp
